/*
   $Id: Bank.h 672 2011-12-09 13:53:53Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __BANK_H__
#define __BANK_H__

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <string>
#include <vector>
using namespace std;

#include "ptypes.h"
#include "buffer.h"
#include "masque.h"
#include "BlackList.h"
#include "Parameters.h"
#include "SeqInfo.h"

class BankBuffer;

/**
 * \brief Bank modelizes the bank. It is a singleton (there is only ONE bank)
 * \note  Bank is used for indexation purpose AND for searching seeds.
 */
class IndexTree;
class Bank: public Singleton<Bank>
{
public:

	friend class Singleton<Bank>;
	void Index(bool memory=false);

private:
	Bank();

public:
    /* Accessors */
    string GetName() const {return __bank_name;};
    const CompressedFileName& GetFileName() const {return __filename;};
    Buffer * GetBiBfr() const {return bi_bfr.get();};
    Buffer * GetFaBfr() const {return GetBiBfr()==NULL ? NULL : fa_bfr.get();};    // If bi_bfr is NULL, we return NULL also

	uint8_t GetNbPass() const { return __nb_pass;};
    uint32_t GetNbWords() const {return Words;};
    uint64_t GetNbSeq() const {return nb_seq;};
    uint32_t GetNbPos() const {return Positions;};
    const SeqInfo& GetSeqInfo() const {return seq_info;};

	uint16_t GetComplexity() const { return complexity;};
	
    uint32_t GetLength() const {return Length;};
    uint32_t GetSequenceLength() const {return SequenceLength;};
    ssize_t GetSize() const {return Size;};
    uint32_t GetPositions() const {return Positions;};
    int GetNodeWidth(int level) const {return node_width[level];};

    SequencePosition GetTrueSeqPos(uint32_t) const;    // Given a position in the Bank, return the true corresponding SequencePosition
	
	/* Giving access to the black list */
	const BlackList& GetBlackList() const { return *blist_aptr; };
		
	/* Some mutators */
	
	uint32_t IncrNbWords() { return ++Words; };
	uint32_t DecrNbWords() { return --Words; };
	
    // After execution of SetBfr, the object is the new owner of bi_bfr and of fa_bfr - The original owner can be destroyed
    void SetBfr(Buffer_ptr bi, Buffer_ptr fa) {bi_bfr=bi; fa_bfr=fa;};

    //void SetLength(uint32_t l) { Length=l;};
    void SetSequenceLength(uint32_t, uint32_t);
    void PushSeq(const string& seq_name, size_t start, size_t end) {seq_info.push_back(SequencePosition(start,end,seq_name)); nb_seq++;};

    friend istream& operator>>(istream& is, Bank&);
    friend ostream & operator<<(ostream&, const Bank&);
    friend class IndexTree;

private:
    Parameters& __parameters;
    Mask&       __mask;         /* The mask used for indexation */
    void        __init_node_width()     // used by the constructors
    {
        //node_width[1]=5;    // sons + shift = 1 + 4
		node_width[1]=1+sizeof(RootLevelOffsetType);    // sons + shift = 1 + 8
        node_width[2]=6;    // hash_code + sons + shift = 1 + 1 + 4
        for(int i=3;i<=MAX_LEAF_LEVEL;i++) node_width[i]=4; // hash_code + sons + shift = 1 + 1 + 2
    };
    void __InitBank(const string&);

    string             __bank_name;      /* Bank name (output display purpose)  */
    CompressedFileName __filename;  /* Bank filename */
    
    uint8_t __nb_pass;     /* The number of passes when  indexing */
	void __SetNbPass();  /* Set the number of passes */

    uint32_t Words;      /* nb of different words in the tree  */
    uint32_t Length;     /* Length (in nucleotides) of the Bank   */
    uint32_t SequenceLength;   /* Length (in nucleotides) of the Bank  sequence (shorter than the Length because here the separators are not counted) */
    size_t Size;      /* size in bytes of the compressed index .bind.bi */
    uint32_t Positions;  /* Total number of positions stored in the tree */
    uint64_t nb_seq ;    /* number of sequences of the Bank (multifasta)*/
	SeqInfo seq_info; // correspondance between positions in the Bank (i.e in .bind.bi.fa) and the true positions in the sequences
    int node_width[MAX_LEAF_LEVEL+1];   /*The width, in bytes, for the node/leaf info at each mask level */

    uint16_t complexity;   /* The complexity algorithm used for indexation (0/1/2) */

    Buffer_ptr bi_bfr; /* The binary index is stored in a Buffer in auto_index mode */
    Buffer_ptr fa_bfr;
    
    auto_ptr<BlackList> blist_aptr;
    IndexTree* tree_ptr;
};

ostream & operator<<(ostream&, const Bank &);
istream& operator>>(istream& is, Bank&);

#endif
