/*
   $Id: CompactTree.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __COMPACTTREE_H__
#define __COMPACTTREE_H__

#include "GlintEngine.h"
#include <map>

#define MAX_LIGNE_SIZE 1000

/**
 * \brief The indexation tree is stored in memory using this class, when used for retrieving data (called by glint)
 *        - CompactTree is a singleton
	      - Retrieving the top nodes: all those functions return a char*, ie an address in the index memory area. 
	        This address should point the a node definition
	
	        - GetTopXXX are used to retrieve the main phyla (top nodes)
	          - GetTop:       parameter = hash_code    - return the corresponding node
	          - GetTopNbSons: parameter = hash_code    - return the number of sons this top node has, top node retrieved from the code
	          - GetTopNbSons: parameter = const char * - return the number of sons this top node has, top node pointed to by parameter
	          - GetTopSons:   parameter = hash_code    - return the first son this top node has, top node retrieved from the code
	          - GetTopSons:   parameter = const char * - return the first son this top node has, top node pointed to by parameter
	
	        - GetXXX are used to retrieve internal nodes or leafs - Note that the leafs are treated exactly like the nodes. 
	          The "sons" of the leaves are the positions, and the positions do not have any son.
	          - GetCode:      parameter = const char * - return the hash code of this node
	          - GetNbSons:    parameter = const char * - return the number of sons this node has
	          - GetSons:      parameters= const char *, int - return the first son this node has, the function called depends on the level
	          - GetBrother:   parameter = const char *, int - return the next brother of this node, the shift depends on the level
	          - GetPos:       parameter = const char * - return the position pointed to by p
*/

class QueryBuffer;
class TreePath;
class CompactTree: public Singleton<CompactTree> {
	public:
		friend class Singleton<CompactTree>;
	    virtual ~CompactTree();
	
    private:
	    CompactTree();
	public:
	    const unsigned char* GetTop(uint32_t code) const {return index - code*_bank.GetNodeWidth(ROOT_LEVEL);};
	    uint8_t GetTopNbSons(uint32_t code) const {return *(GetTop(code));};
	    uint8_t GetTopNbSons(const unsigned char* p) const {return *p;};
	    const unsigned char* GetTopSons(uint32_t code) const    { return base_index + __Ptr[ROOT_LEVEL](GetTop(code)+1);}
	    const unsigned char* GetTopSons(const unsigned char* p) const {return base_index + __Ptr[ROOT_LEVEL](p+1);};
	
	    inline uint8_t GetCode(const unsigned char* p) const {return p[0];};   // Return the code, assuming p is a correctly initialized ptr
	    uint8_t GetNbSons(const unsigned char* p) const {return p[1];};        // Return the number of sons, assuming p is a correctly initialized ptr
	    const unsigned char*  GetSons(const unsigned char *p, int level) const {return p-__Ptr[level](p+2);};
	    const unsigned char*  GetBrother(const unsigned char* p, int level) const {return p-_bank.GetNodeWidth(level);};
	    uint32_t GetPos(const unsigned char* p) const {return *(uint32_t*)p;};
	    const unsigned char*  GetIndex(uint32_t code) const    { return base_index;};
		
		uint32_t GetNbPos() { return NbPos;};
	
	    const unsigned char* GetSeq() const {return base_seq;};
	    unsigned char Getc(uint32_t i) const {return base_seq[i];};
	
	    size_t GetSequenceLength() const {return seq_length;};
	    size_t GetIndexSize() const { return index_size; };

	    friend ostream & operator<<(ostream&, const CompactTree&);
	
		void CompareWindow(const Window&, uint32_t qpos, GlintEngineSingleStrand&, const QueryBuffer&) const;

	protected:
	    const Parameters& _parameters;
   	    const Bank & _bank; 
	    const Mask& _mask;

	private:
	    void __InitFromBfr();            // Called by the constructor
	    void __InitFromFiles();          // Called by the constructor
	    const unsigned char* base_index; // The base address of the allocated memory for the binary index
	    const unsigned char* index;      // The start point (near the end of the area)
	    size_t index_size;               // The size of the index buffer
	    const unsigned char* base_seq;   // The base address of the allocated memory for the sequence
	    const unsigned long nb_top;      // The number of top nodes
	    size_t seq_length;               // The size of the sequence buffer
	    bool InitedFromBuffer; 			 // True if the dababase has been inited from the buffer
	    RootLevelOffsetType (*__Ptr[MAX_LEAF_LEVEL+1])(const unsigned char*);   // An array of functions, initialized by the constructor
	    
	    mutable uint32_t NbPos;             
	    
	
	
	    int __Search(const Window &, TreePath&) const;
	    int __Search(const Window&, const unsigned char*, uint8_t, uint8_t, TreePath&) const;
	    int __ReverseSearch(const Window &, TreePath&) const;
	    int __ReverseSearch(const Window&, const unsigned char*, uint8_t, uint8_t, TreePath&) const;
	    const unsigned char * __HorizSearch(uint32_t, const unsigned char*, uint8_t, uint8_t) const;
	    void __RetrievePos(const unsigned char *, uint8_t, TreePath&, uint32_t, GlintEngineSingleStrand&, const QueryBuffer&) const;
	    virtual void __RetrievePosTab(const unsigned char *, uint8_t, uint8_t, uint32_t, GlintEngineSingleStrand&, const QueryBuffer&) const;
	
	    static RootLevelOffsetType __Ptr2(const unsigned char *p) {return (RootLevelOffsetType)(*((uint16_t *)p));};
	    static RootLevelOffsetType __Ptr4(const unsigned char *p) {return (RootLevelOffsetType)(*((uint32_t *)p));};
	    static RootLevelOffsetType __Ptr8(const unsigned char *p) {return (*((RootLevelOffsetType *)p));};
	
		// the following useful for glintstat		
		ostream & __PrintPositions (ostream &) const;
		ostream & __PrintFrequency (ostream &) const;
	    void __PrintBranch(ostream&, const unsigned char*, uint8_t, uint8_t, const string&, int) const;
   	    void __PrintPhylum(ostream&,uint32_t) const;
   	    int  __FreqPhylum (multimap<int,string>&, uint32_t code) const;
   	    int  __FreqBranch (multimap<int,string>&, const unsigned char*, uint8_t, uint8_t, const string&) const;
};

ostream & operator<<(ostream&, const CompactTree&);

/*
 * CompactTreeSelf RETIRE car pas possible d'heriter d'un singleton (des miens en tous cas) !!!
 * 
class CompactTreeSelf: public CompactTree {
	public:
		virtual void __RetrievePosTab(const unsigned char *, uint8_t, uint32_t,MatchPile&) const;

};
*/
#endif
