/*
   $Id: buffer.cpp 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <string.h>
#include "ptypes.h"
#include "buffer.h"
#include "SeqInfo.h"

/**
 * \brief 1st constructor: min,max
 *        - When the buffer is built using this constructor, it is responsible for the memory allocation/free:
 *              - the array is allocated when necessary (it at first Put() call)
 *              - the array may be reallocated if necessary (when Put cannot be executed any more)
 *              - Put, Clear are available
 */
Buffer::Buffer(size_t min,size_t max): min_capacity(min?min:1),max_capacity(max),cur_capacity(0),begin(NULL),end(NULL),owner(true){}

/**
 * \brief 2nd constructor: begin, size
 *        - When the buffer is built using this constructor, it is used to wrap an already allocated memory array.
 *              - The buffer is not the owner of the data: no alloc,realloc, free may occur
 *              - no Put, no clear is available
 *              - max_capacity is set to the array size
 */

Buffer::Buffer(void* b, size_t s): min_capacity(0),max_capacity(size()),cur_capacity(size()),begin(b),end((char*)b+s),owner(false){}

/**
 * \brief Copy constructor
 *        - When the buffer is built using this constructor, the owner/not owner status is copied from the source buffer
 */
Buffer::Buffer(const Buffer& bfr): 
	min_capacity(bfr.min_capacity),
	max_capacity(bfr.max_capacity),
	cur_capacity(bfr.cur_capacity),
	begin(bfr.begin),
	end(bfr.end),
	owner(bfr.owner)
{
	// If the object is responsible for the memory allocation/free, we have to create a NEW area and copy the data
	if (owner && begin != NULL)
	{
		begin = malloc(cur_capacity);
        if (begin==NULL)
			throw(bad_alloc());
        else
        {
			size_t size = (uint8_t*) bfr.end - (uint8_t*)bfr.begin;
	 		memcpy(begin, bfr.begin, size);
	 		end = (void*)((uint8_t*)begin + size);
		}
	}
	//cerr << "copy constructor of Buffer " << begin << " " << cur_capacity << " " << owner << endl;
}

/**
 * \brief *2 the buffer capacity, and realloc memory. Called by Put when the current buffer is full
 * \return true if success, false if the max capacity was already reached
 * \exception bad_alloc
 */
bool Buffer::__Realloc() const throw(bad_alloc)
{
    if (begin==NULL)
    {
        cur_capacity = min_capacity;
        begin = malloc(cur_capacity);
        if (begin==NULL)
			throw(bad_alloc());
        else
        {
            end = begin;
            //cout << "Buffer::_Realloc - Initial buffer " << cur_capacity << '\n';
        }
    }
    else
    {
        size_t wanted_capacity = 2*cur_capacity;
        size_t old_size = size();

        if (max_capacity==0 || wanted_capacity<=max_capacity)
        {
            void* temp = realloc(begin, wanted_capacity);
            if (temp==NULL)
                throw(bad_alloc());
            else
            {
                begin = temp;
                end = (void*)((unsigned char*)temp + old_size);
                cur_capacity = wanted_capacity;
                //cerr << "Buffer::_Realloc - realloc now capacity " << cur_capacity << "\n";
            }
        }
        else
            return false;
    }
	return true;
}

void Buffer::Reverse()
{
	if(owner) 
		Reverse1((unsigned char*)GetBase(),(uint32_t)size());
}
