/*
   $Id: ptypes.h 793 2014-08-21 07:32:12Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

/* Personal types for compatibility */


#ifndef _PTYPES_H
#define _PTYPES_H

#include <cstdlib>

#include "rtstack.h"
#include <stdint.h>

typedef long DiagInt;

enum EXIT
	{
	EXIT_Success = 0,
	EXIT_NotStarted = 1,
	EXIT_FatalError = 2,
	EXIT_Except = 3
	};

enum STRAND
	{
	WATSON = 0,
	CRICK = 1
	};

typedef enum orient {Minus, Plus} OrientType;

#define MAX_FASTA_HEADER_SIZE 1000                    /* The max number of characters allowed for the header line of the fasta files */
#define QBUFFER_INITIAL_SIZE  2000000                 /* The initial size of QBuffer, we realloc if necessary */
#define SEQUENCE_SEPARATOR_SIZE 10

#define SBUFFER_INITIAL_SIZE   62500               /* The initial size of Sbuffer (used for indexation), we realloc if necessary */
#define SBUFFER_MAX_SIZE       20000000                /* The maximal size of Sbuffer, we dump the sbuffer if necessary */
/*#define SBUFFER_MAX_SIZE       4000000000*/                 /* The maximal size of Sbuffer, we dump the sbuffer if necessary */

#define DEFAULT_SLICE_SIZE 10000000
#define DEFAULT_TANDEM_SIZE 10000000000

#define SMALL_BUFFER_SIZE 255

/* Some data structure dimensions and other parameters */
#define BLACK_LIST_SIZE 60000       /* The number of slots in the black list */
#define HSP_PILE_SIZE   50000      	/* The size of an hsp pile (number of hsp) */
#define MATCH_PILE_SIZE 800000      /* The size of a MatchPile (number of matches) */

#define ROOT_LEVEL 1                /* A lot of things will hang if ROOT_LEVEL != 1 */
#define MAX_LEAF_LEVEL 8            /* Should be possible (but useless) to change this value */

/*A simple region structure */
struct regionFreq
{
	uint32_t 			start;
	uint32_t 			end;
};

//Pair-end specific constants
#define DEFAULT_MAX_DISTMATE 10000   /* Default Maximum allowed distance between pair-end reads */

enum FRAGMENTS {
		FWD = 1,
		BWD = 2
	};

#define DEFAULT_QUALITY 65
	
#define NB_BEST_HITS "X0"
#define NB_SUBOPT_HITS "X1"
#define SUBOPTALIGN_SCORE "XS"
#define NB_BEST_PAIRHITS  "X2"
#define NB_SUBOPT_PAIRHITS "X3"

//#define ALIGN_CONSTANTS_DEF_PRMS 1,-3,-1,-5,-2,15,25  /* AlignConstants constructor parameters in the following order : match, mismatch, mismatch(match) with an N, gap opening penalty, gap extension penalty, dropoff for ungapped aligments, cutoff for gapped alignments  */
#define ALIGN_CONSTANTS_MATCH 1
#define ALIGN_CONSTANTS_MISMATCH -3
#define ALIGN_CONSTANTS_NMATCH -1
#define ALIGN_CONSTANTS_GAPOPEN -5
#define ALIGN_CONSTANTS_GAPEXTEN -2
#define ALIGN_CONSTANTS_UNGAP_DROPOFF 15
#define ALIGN_CONSTANTS_GAP_DROPOFF 25
#define DYNPROG_MAT_SIZE 1000           /* The size of the matrices used by dynamic programming */
#define DISPLAY_LINE_SIZE 60

/* The complexity management */

#define COMPLEXITY_NONE          0
#define COMPLEXITY_DUST          1
#define DEFAULT_COMPLEXITY       COMPLEXITY_DUST

/* glintindex */
#define END_OF_SEQUENCE          '$'
#define DEFAULT_MAX_COUNT        100
//#define DEFAULT_MASK             "13112411211341"    /* The default Mask is now of 14 letters  */
//#define DEFAULT_MASK             "13112111211311"    /* The default Mask is now of 14 letters  */
#define DEFAULT_MASK             "130110211012110311"    /* The default Mask is now of 14 letters but of size 19 (0 are not counted)*/
#define FA_EXTENSION             "fa"
#define BI_EXTENSION             "bi"
#define HE_EXTENSION             "he"
#define BL_EXTENSION             "bl"

enum AlignStatus {ALIGNED,GAPED};

/*                          The different output formats
 *
 *   M0 : dummy format (no output, useful to measure pure computations performance)
 *   M1 : standard output format including a description and the full alignment
 *   M2 : a simple tabular output with the following header line:
 *         #query\tq_start\tq_end\tsubject\ts_start\ts_end\tstrand\tscore\tsize
 *   M3 : an improved tabular format with an editing script
 *   M4 : a format dedicated to analysis of the output
 *   M8 : a blast-like -m 8 output format (without header): a tab separated tabular ouput with the following fields (in thart order)
 *       q_name s_name pident size #mismatch #gaps  q_start q_end s_start s_end "1e-0" score
 *   M9 : a blast-like -m 9 output format (same as M8 but with a header)
 */

#define OUTPUT_FORMAT_UNKNOWN 255
#define OUTPUT_FORMAT_ILLEGAL 254
#define OUTPUT_FORMAT_M0      0
#define OUTPUT_FORMAT_M1      1
#define OUTPUT_FORMAT_M2      2
#define OUTPUT_FORMAT_M3      3
#define OUTPUT_FORMAT_M4      4
#define OUTPUT_FORMAT_M5      5
#define OUTPUT_FORMAT_M8      8
#define OUTPUT_FORMAT_M9      9
#define OUTPUT_FORMAT_FA     13
#define OUTPUT_FORMAT_FASTQ  14
#define OUTPUT_FORMAT_SAM    15
#define OUTPUT_FORMAT_BAM    16
#define OUTPUT_FORMAT_SAMEXT 17
#define OUTPUT_FORMAT_BAMEXT 18
#define OUTPUT_FORMAT_BED    19

#define DEFAULT_GLINT_OUTPUT_FORMAT            OUTPUT_FORMAT_M1
#define DEFAULT_GLINT_OUTPUT_FORMAT_DSP        "1"
#define DEFAULT_GLINTCONVERT_OUTPUT_FORMAT     OUTPUT_FORMAT_FASTQ
#define DEFAULT_GLINTCONVERT_OUTPUT_FORMAT_DSP "fastq"

#define M2_HEADER "#q_name\tq_start\tq_end\ts_name\ts_start\ts_end\tstrand\tscore\tsize"
#define M3_HEADER "#query\tsubject\tescript([q_start-q_end]->[s_start-s_end](size,strand,score,evalue,pident)"
#define M5_HEADER "query\tqgaps\tqstart\tqend\tqsize\tsubject\tsgaps\tsstart\tsend\tssize\tmismatch\tpident\tscore\tstrand"

#define PLUS 1
#define MINUS -1

#define OUTPUT_PLUS  "Plus/Plus"
#define OUTPUT_MINUS "Plus/Minus"

#define MAX_WARNINGS 100

/*                           The different input formats
 * 
 *  fa : fasta format
 *  sam: sam/bam format
 *  bam: sam/bam format
 */

#define INPUT_FORMAT_UNKNOWN 255
#define INPUT_FORMAT_FA    13
#define INPUT_FORMAT_FASTQ 14
#define INPUT_FORMAT_SAM   15
#define INPUT_FORMAT_BAM   16

#define DEFAULT_GLINT_INPUT_FORMAT            INPUT_FORMAT_FA
#define DEFAULT_GLINT_INPUT_FORMAT_DSP        "fasta"
#define DEFAULT_GLINTCONVERT_INPUT_FORMAT     INPUT_FORMAT_BAM
#define DEFAULT_GLINTCONVERT_INPUT_FORMAT_DSP "bam"

/*
    Default values for Compare_Parameters
*/

#define DEFAULT_STEP 1

#define DEFAULT_CUTOFF_SLIDE  20
#define DEFAULT_CUTOFF_DYN    25
#define DEFAULT_CUTOFF_HSP    30
#define DEFAULT_LOWER_LEVEL    2
#define DEFAULT_DEEPEST_LEVEL  2

#define DEFAULT_BLASTZ_BEHAVIOR 0

/*
    Default values for Linking option of compare
*/

#define MAX_REPEATS 1000
#define DEFAULT_MAX_CHAIN_OVERLAP 0.05 //no more than 5% overlap
#define REPEAT_THRESHOLD_SIZE 1000

#define MAX_OVERLAP_HSP 100

typedef uint64_t RootLevelOffsetType;

/*
 *  Supporting the compressed files	
 *  Highest compression factor first (this is important for SearchBestCompression() )
 */
enum CompressionAlgorithm
{
	NOCOMPRESSION,
	ANYCOMPRESSION,
	XZ,
	LZMA,
	BZ2,
	GZ,
	Z
};
#define NB_COMPRESSIONS 7

/*
    Classe SLogger = If not in DEBUG, use the trivial logger for speed
*/

#include "logger.h"
#ifdef DEBUGLOG
   typedef SLogger RealLogger;
#else
   typedef STrivialLogger RealLogger;
#endif

/**
 * \brief The classes who declare a private inheritance from this class cannot redefine any operator=, nor any copy constructor. 
 *        Thus, they are not copyable
 * 
 * \par How to create a class which should not be copied ?
 * Simply as follows:
 * \code
class my_class: private NonCopyable {
...
}
* \endcode
*/

class NonCopyable
{
    public:
	    NonCopyable(){};
    private:
	    NonCopyable(const NonCopyable&) {};
	    void operator=(const NonCopyable&) {};
};

/** 
 * \brief A light implementation of the Singleton pattern
 *        - Singletons defined with this code have the same lifetime as the program itself
 *        - The objects are destroyed when the program leaves
 *        - The destruction occurs in the reverse order of memory allocation, thanks to the atexit function
 * \par How to create a singleton from a class called my_class ?
 *     -# my_class must extend the class Singleton<my_class>
 *     -# Singleton<my_class> must be a friend of my_class
 *     -# The constructor of my_class must be declared as private
 * Here is some complete code:
 * \code
class my_single: public Singleton<my_single> {
	private:
	   my_single();
	public:
	   friend class Singleton<my_single>;
};
 * \endcode
 */

/*class SingletonInitializer
{
	public:
		SingletonInitializer(){};  // default constructor
		bool GetStatus() const { return __initialized; };
	
	private:
		bool __initialized;
};
extern SingletonInitializer __INIT;
*/ 
template<typename T> class Singleton: private NonCopyable
{
	protected:
	    Singleton () {};
	    virtual ~Singleton(){};

    public:
	    static T* InstancePtr ()
	    {
	        if (self!=NULL)
	        {
		        return static_cast<T*> (self);
			}
			else
			{
				throw(rtstack("ERROR - THE SINGLETON SHOULD BE ALREADY INITIALIZED"));
			}
			return static_cast<T*> (self);
	    };
	    static void Init ()
	    {
	        if (self==NULL)
	        {
	            self = new T;
	            atexit(Singleton<T>::Remove);
			}
	    };
	    static void ReInit()
	    {
			Remove();
			Init();
		}
	    static T& Instance ()
	    {
			return *InstancePtr();
		};
		static void Remove() 
		{ 
			if (self!=NULL) 
				delete self; 
			self = NULL;			
		};

    private:
	    static T* self;
};

template <typename T> T* Singleton<T>::self = NULL;
#endif
