/*
   $Id: utils.h 804 2016-03-01 15:15:56Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __UTILS_H__
#define __UTILS_H__

#include "glint.h"

#include <stdexcept>
#include <stdio.h>
#include <execinfo.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <memory>
#include <string>
#include <vector>
#include <algorithm>
#include "pstream.h"
#include "Parameters.h"

using namespace std;

/*
 * some conversion functions of general use
 */

uint32_t String2uint32_t(const string& value);
uint16_t String2uint16_t(const string& value);
char String2Char(const string& value);
inline string ToUpper(const string& s) {string su=s;for(unsigned int i=0;i<su.length();++i)su[i]=toupper(su[i]);return su;}
inline string ToLower(const string& s) {string su=s;for(unsigned int i=0;i<su.length();++i)su[i]=tolower(su[i]);return su;}

// Discovering the sequence name from an already opened file
string GetSeqName(FILE *fp);
// Getting the name of the read:
//  according to the sam format description manual 
//  NGS platforms sequence both ends of an insert, called the read, producing two fragments labelled ReadName/1 and ReadName/2
string GetReadName(const string& fragmentId );

// Print a header
void PrintHeader();

// Glint version 
string GetGlintVersion();

/* 
 * MAlloc: call malloc and throw an exception in case of error
 * ReAlloc:call realloc and throw an exception in case of error
 * FOpen : Take a filename (1st parameter), append the extension if specified, specify the open mode, and maybe a compression algorithm
 *         The FILE* may be stdin or stdout if the filename is "-"
 *         Return a FILE* or throw an exception
 * FOpen : Take a CompressedFileName, specify the open mode and call the first version of FOpen, specifying the "" extension
 * FClose: close the FILE* and set it to NULL, so that it will be simple knowing if the file is opened or not
 * IstreamPtr: same as istream, but maybe using a compression algorithm
 * OstreamPtr: same as ostream, but maybe using a compression algorithm
 */
void* MAlloc(size_t size);
void* ReAlloc(void*, size_t);
class IstreamPtr;
class OstreamPtr;
class CompressedFileName;

inline void  FClose(FILE*& fp) { if (fp!=NULL && fp!=stdout && fp!=stdin){fclose(fp); fp=NULL;} }
FILE* FOpen(const string&, const string&, const char*, CompressionAlgorithm compression = NOCOMPRESSION );
FILE* FOpen(const CompressedFileName& cfn, const char* mode);

/**
 * \brief An automatic pointer to an ifstream, to an ipstream (to manage compression), or a pointer to cin
 */
class IstreamPtr
{
	public:
		IstreamPtr(const string& name, const string& extension, CompressionAlgorithm compression=NOCOMPRESSION, ios_base::openmode mode = ios_base::in);
		IstreamPtr(const CompressedFileName& name, ios_base::openmode mode = ios_base::in);
		istream& operator*() 
		{ 
			if ( ifstream_aptr.get() != NULL )
			{
				return *ifstream_aptr;
			}
			else if ( ipstream_aptr.get() != NULL )
			{
				return *ipstream_aptr;
			}
			else
			{
				return cin;
			}
		};
		istream* operator->() 
		{
			if ( ifstream_aptr.get() != NULL )
			{
				return ifstream_aptr.get();
			}
			else if ( ipstream_aptr.get() != NULL )
			{
				return ipstream_aptr.get();
			}
			else
			{
				return &cin;
			}
		};
		
	private:
		void __Init(const string&, const string&, CompressionAlgorithm, ios_base::openmode mode);
		auto_ptr<ifstream> ifstream_aptr;
		auto_ptr<redi::ipstream> ipstream_aptr;
};

/**
 * \brief An automatic pointer to an ofstream, to an opstream (to manage compression), or a pointer to cout
 */
class OstreamPtr
{
	public:
		OstreamPtr (const string&, const string&, CompressionAlgorithm compression=NOCOMPRESSION, ios_base::openmode mode = ios_base::out);
		ostream& operator*() 
		{ 
			if ( ofstream_aptr.get() != NULL )
			{
				return *ofstream_aptr;
			}
			else if ( opstream_aptr.get() != NULL )
			{
				return *opstream_aptr;
			}
			else
			{
				return cout;
			}
		};
		ostream* operator->() 
		{
			if ( ofstream_aptr.get() != NULL )
			{
				return ofstream_aptr.get();
			}
			else if ( opstream_aptr.get() != NULL )
			{
				return opstream_aptr.get();
			}
			else
			{
				return &cout;
			}
		};
		
	private:
		auto_ptr<ofstream> ofstream_aptr;
		auto_ptr<redi::opstream> opstream_aptr;
};
	
//void bin_print(long n, int taille);
//void bin_print_char(char n);
//void bin_print_short(short n);
//void bin_print_int(int n);
//void bin_print_unlong(unsigned long n);

/*
 * Used by glintchain
 */

bool ValidCutoffParam( const string cutoffs );
bool Validuint32_t(const string& s, bool failIfLeftoverChars = true);
bool ValidCovar(const string& s, bool failIfLeftoverChars = true);

#endif
